;;;                  -*- Mode: Lisp; Package: GEOMETRY -*-                  ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         geometry.lsp                                              ;;;
;;; Description:  BOOMS geometrical functions                               ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         21-9-96                                                   ;;;
;;; Package:      GEOMETRY                                                  ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  G E O M E T R Y  file                                                  ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :geometry)


;;;-------------------------------------------------------------------------;;;
;;;  D E S T R U C T I V E  function versions                               ;;;
;;;-------------------------------------------------------------------------;;;
;;; These are destructive-efficient versions of the box corner accessors.

(defun NBOX-TOP-LEFT (pos box)
  (nmake-position pos (box-left  box) (box-top    box)))

(defun NBOX-TOP-RIGHT (pos box)
  (nmake-position pos (box-right box) (box-top    box)))

(defun NBOX-BOTTOM-RIGHT (pos box)
  (nmake-position pos (box-right box) (box-bottom box)))

(defun NBOX-BOTTOM-LEFT (pos box)
  (nmake-position pos (box-left  box) (box-bottom box)))

;;; These are the box accessors efficiently returning the same object - they
;;; are named with a prefix of NN.
;;; Carfeul: do not use two calls of the same function in a single statment.

(defun NNBOX-TOP-LEFT (box)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos (box-left  box) (box-top    box))))

(defun NNBOX-TOP-RIGHT (box)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos (box-right box) (box-top    box))))

(defun NNBOX-BOTTOM-RIGHT (box)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos (box-right box) (box-bottom box))))

(defun NNBOX-BOTTOM-LEFT (box)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos (box-left  box) (box-bottom box))))


;;;-------------------------------------------------------------------------;;;
;;;  B O X E S - I N T E R S E C T - P  definition                          ;;;
;;;-------------------------------------------------------------------------;;;

;;; Lower-right corner of a x1,y1,x2,y2 rectangle.  NN means always return the
;;; same position object.
(defun NNPOSITION-MAX (pos1 pos2)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos
                    (max (position-x pos1) (position-x pos2))
                    (max (position-y pos1) (position-y pos2)))))

;;; Upper-left corner of a x1,y1,x2,y2 rectangle.
(defun NNPOSITION-MIN (pos1 pos2)
  (let ((pos #.(make-position 0 0)))
    (nmake-position pos
                    (min (position-x pos1) (position-x pos2))
                    (min (position-y pos1) (position-y pos2)))))

;;; pos1 is left and above of pos2.
(defun POSITION<= (pos1 pos2)
  (and (<= (position-x pos1) (position-x pos2))
       (<= (position-y pos1) (position-y pos2))))

;;; Return a normalized box - upper-left,lower-right.  Return box if already
;;; normalized.
(defun GET-NORMAL-BOX (box)
  (let ((p1 (nnbox-top-left box)) (p2 (nnbox-bottom-right box)))
    (if (position<= p1 p2)
      box
      (make-box-from-corners (nnposition-min p1 p2) (nnposition-max p1 p2)))))

;;; Normalize a box destructively.  Dont touch if already normalized.
(defun NORMALIZE-BOX (box)
  (let ((p1 (nnbox-top-left box)) (p2 (nnbox-bottom-right box)))
    (if (position<= p1 p2)
      box
      (nmake-box-from-corners
       box (nnposition-min p1 p2) (nnposition-max p1 p2)))))

;;; Do two boxes intersect?
(defun BOXES-INTERSECT-P (box1 box2)
  (box-intersect-p (get-normal-box box1) (get-normal-box box2)))

;;; Do two boxes intersect, destructively normalizing both?
(defun NBOXES-INTERSECT-P (box1 box2)
  (box-intersect-p (normalize-box box1) (normalize-box box2)))


;;;-------------------------------------------------------------------------;;;
;;;  B O X - L I N E - I N T E R S E C T - P  definition                    ;;;
;;;-------------------------------------------------------------------------;;;
;;; This is actually not used in BOOMS, but kept for fun.

;;; The cross-product of two points.
(defun POSITION-PRODUCT (pos1 pos2)
  (- (* (position-x pos1) (position-y pos2))
     (* (position-x pos2) (position-y pos1))))

;;; Is a given point p to the right of a line p1-p2?  returns 1/0/-1.
(defun RIGHT-OF (p p1 p2)
  (let ((ptmp  #.(make-position 0 0))
        (p2tmp #.(make-position 0 0)))
    (ncopy-position ptmp  p)
    (ncopy-position p2tmp p2)
    (signum (position-product (nposition- p2tmp p1) (nposition- ptmp p1)))))

;;; Does a line p1-p2 straddle a line q1-q2?
(defun STRADDLE (p1 p2 q1 q2)
  (< (* (right-of q1 p1 p2) (right-of q2 p1 p2)) 1))

;;; Does the two lines p1-p2 and q1-q2 intersect?
(defun LINES-INTERSECT (p1 p2 q1 q2)
  (and (<= (min (position-x p1) (position-x p2))
           (max (position-x q1) (position-x q2)))
       (<= (min (position-x q1) (position-x q2))
           (max (position-x p1) (position-x p2)))
       (<= (min (position-y p1) (position-y p2))
           (max (position-y q1) (position-y q2)))
       (<= (min (position-y q1) (position-y q2))
           (max (position-y p1) (position-y p2)))
       (straddle p1 p2 q1 q2)
       (straddle q1 q2 p1 p2)))

;;; Does a line intersect with a box?
(defun BOX-LINE-INTERSECT-P (box p1 p2)
  (or (inside-box-p p1 box)
      (inside-box-p p2 box)
      (lines-intersect p1 p2 (nnbox-top-left  box) (nnbox-bottom-right box))
      (lines-intersect p1 p2 (nnbox-top-right box) (nnbox-bottom-left  box))))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :GEOMETRY)

