;;;                    -*- Mode: Lisp; Package: MISC -*-                    ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         gr-misc.lsp                                               ;;;
;;; Description:  BOOMS miscellaneous graphic functions                     ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         24-12-96                                                  ;;;
;;; Package:      MISC                                                      ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  G R - M I S C  file                                                    ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :misc)


;;;-------------------------------------------------------------------------;;;
;;;  M I S C  definitions                                                   ;;;
;;;-------------------------------------------------------------------------;;;

;;; This macro is used to disable interruption of one user-function by another
;;; using some lock variable.
(defmacro WITH-LOCK (lock-var &body body)
  `(unless ,lock-var (let ((,lock-var t)) ,@body)))

;;; Macro for handling no events,used for document windows.
(defmacro WITH-EVENTS-DISABLED ((&rest windows) &body body)
  `(unwind-protect
       (progn
         ,@(mapcar #'(lambda (w) `(enable-events ,w nil)) windows)
         ,@body)
     ,@(mapcar #'(lambda (w) `(enable-events ,w t)) windows)))

;;; Macro to use a clip box which will be intersected by nested activations.
;;; Assume nested activations are always on the same window.

(defvar *CLIPPING-BOX* nil)

(defmacro WITH-CLIPPING ((window box) &body body)
  `(unwind-protect
       (let ((*clipping-box* (cond
                               ((not *clipping-box*) ,box)
                               ((box-intersect-p *clipping-box* ,box)
                                (box-intersect *clipping-box* ,box))
                               (t #.(make-box 0 0 0 0)))))
         (set-clipping-box ,window *clipping-box*)
         ,@body)
     (set-clipping-box ,window *clipping-box*)))

;;; A macro to set a temporary text background mode.
(defmacro WITH-BACKGROUND-MODE ((window mode) &body body)
  `(let ((old-mode (background-mode ,window)))
     (unwind-protect
         (progn
           (setf (background-mode ,window) ,mode)
           ,@body)
       (setf (background-mode ,window) old-mode))))

;;; Get the windows caption-area height.
(defun GET-CAPTION-HEIGHT ()
  (win:getsystemmetrics win:sm_cycaption))

;;; Get the screen dimensions - this will work and take into account the
;;; taskbar area, assuming its on the bottom of the screen.

(defun GET-SCREEN-SIZE-X ()
  (win:getsystemmetrics win:sm_cxfullscreen))

(defun GET-SCREEN-SIZE-Y ()
  (+ (win:getsystemmetrics win:sm_cyfullscreen) (get-caption-height)))

;;; Closing a menu and its submenus recursively.

(defmethod MENU-CLOSE ((x t))
  nil)

(defmethod MENU-CLOSE ((items list))
  (dolist (item items)
    (menu-close item)))

(defmethod MENU-CLOSE ((mi menu-item))
  (menu-close (menu-item-value mi)))

(defmethod MENU-CLOSE ((m pc::windows-menu))
  (dolist (mi (menu-items m))
    (menu-close mi))
  (close m))

;;; Make all widgets not draggable/droppable.

(defmethod DRAGGABLE-P ((drag-widget dialog-item))
  nil)

(defmethod DROPPABLE-P ((dragged-over-widget dialog-item)
                        (drag-widget dialog-item))
  nil)

;;; Allowing setting of window-state.
(defun (setf WINDOW-STATE) (new-state window)
  (case new-state
    ((:shrunk :hidden)  (shrink-window window t))
    ((:icon :minimized) (shrink-window window nil))
    (t                  (zoom-window window new-state))))

;;; Openning a whole branch of an outline.
(defun OUTLINE-OPEN-ALL-VALUE (outline value)
  (let ((items
         (list
          (find value (dialog-item-range outline)
                :key #'outline-item-value :test #'equalp))))
    (while items
      (setf (outline-item-state (first items)) :open)
      (setf items
            (append (rest items) (outline-item-subitems (first items)))))))

;;; I replace read-from string by a version that will not crash on something
;;; like x:x with a package error.

(defvar *OLD-READ-FROM-STRING* nil)

;;; This should evaluate when loading.
(eval-when (:load-toplevel :execute)
  (unless *OLD-READ-FROM-STRING*
    (setf *old-read-from-string* #'read-from-string)
    (fmakunbound 'read-from-string)))

(defun READ-FROM-STRING (string &optional (eof-error-p t) eof-value
                                &key (start 0) end preserve-whitespace)
  (multiple-value-bind
      (obj/idx err)
      (ignore-errors
       (multiple-value-list
        (funcall *old-read-from-string*
                 string eof-error-p eof-value :start start :end end
                 :preserve-whitespace preserve-whitespace)))
    (if err eof-value (values-list obj/idx))))


;;;-------------------------------------------------------------------------;;;
;;;  C O L O R - B O X  widget definitions                                  ;;;
;;;-------------------------------------------------------------------------;;;
;;; This is a color box widget - shows a color, and pops the standand color
;;; selection dialog on a mouse click.

;;;; The widget definition.

(defclass COLOR-BOX (pc::lisp-widget) ())

(defclass COLOR-BOX-PANE (pc::lisp-widget-window pc::lisp-widget-top-window)
  ((item :accessor pane-item :initform nil :initarg :dialog-item)))

(defmethod WIDGET-DEVICE ((item color-box) dialog) 'color-box-pane)

(defmethod DEVICE-OPEN ((widget-window color-box-pane) options)
  (call-next-method widget-window (list* :user-scrollable nil options)))

;;; This is the shadow line width for a 3d effect.
(defconstant +SHADOW-WIDTH+ 2)
(defconstant +SHADOW-WIDTH/2+ (round +shadow-width+ 2))

;;; Displaying the widget - the box is full of the selected color, a 3d button
;;; effect and a black border.  This handles a non-rgb value, which will be
;;; drawn as a black cross on a white background.
(defmethod REDISPLAY-WINDOW ((window color-box-pane) &optional box)
  (call-next-method)
  (let* ((value (dialog-item-value (pane-item window)))
         (rgb-p (rgb-p value)))
    (with-background-color (window (if rgb-p value white))
      (clear-page window)
      (let ((top-left     #.(make-position 0 0))
            (top-right    #.(make-position 0 0))
            (bottom-left  #.(make-position 0 0))
            (bottom-right #.(make-position 0 0))
            (visible      (visible-box window)))
        (nbox-top-left     top-left     visible)
        (nbox-top-right    top-right    visible)
        (nbox-bottom-left  bottom-left  visible)
        (nbox-bottom-right bottom-right visible)
        (nposition+
         top-left
         #.(make-position    +shadow-width/2+     +shadow-width/2+))
        (nposition+
         top-right
         #.(make-position (- +shadow-width/2+)    +shadow-width/2+))
        (nposition+
         bottom-left
         #.(make-position    +shadow-width/2+  (- +shadow-width/2+)))
        (nposition+
         bottom-right
         #.(make-position (- +shadow-width/2+) (- +shadow-width/2+)))
        (unless rgb-p
          (with-foreground-color (window black)
            (with-line-width (window 1)
              (draw-line window top-left bottom-right)
              (draw-line window bottom-left top-right))))
        (with-line-width (window +shadow-width+)
          (with-foreground-color (window white)
            (draw-line window top-left bottom-left)
            (draw-line window top-left top-right))
          (with-foreground-color (window gray)
            (draw-line window bottom-right top-right)
            (draw-line window bottom-right bottom-left)))
        (decf (box-right visible))
        (decf (box-bottom visible))
        (with-foreground-color (window black)
          (draw-box window visible)))
      t)))

;;; Redisplay when value is set.
(defmethod WIDGET-SET-VALUE ((window color-box-pane) item new-value
                             old-value recursive-p)
  (unless (eq (window-state window) :shrunk)
    (redisplay-window window))
  t)

;;; Pop up the color selection dialog on a mouse click.
(defmethod EVENT ((window color-box-pane) event buttons data time)
  (select event
    (mouse-left-down
     (let* ((item (pane-item window))
            (color
             (ask-user-for-color
              :initial-color (if (rgb-p (dialog-item-value item))
                               (dialog-item-value item)
                               nil)
              :show-custom-colors t)))
       (when color (set-dialog-item-value item color))))
    (t (call-next-method))))

;;; No drawing/erasing of rectangular focus.

(defmethod PC::LISP-WIDGET-DRAW-FOCUS ((widget-window color-box-pane)
                                       (widget color-box))
  t)

(defmethod PC::LISP-WIDGET-CLEAR-FOCUS ((widget-window color-box-pane)
                                        (widget color-box))
  t)


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :GR-MISC)

