;;;                  -*- Mode: Lisp; Package: NETWORK -*-                   ;;;
;;; ----------------------------------------------------------------------- ;;;
;;; File:         network.lsp                                               ;;;
;;; Description:  BOOMS functional network kernel                           ;;;
;;; Author:       Eli Barzilay - eli@cs.bgu.ac.il                           ;;;
;;; Date:         24-12-96                                                  ;;;
;;; Package:      NETWORK                                                   ;;;
;;; ----------------------------------------------------------------------- ;;;

#|

This file is part of the BOOMS system:
+-----------------------------------------------------------------------------+
|                                                                             |
| /######   /#####   /#####  /##     /##  /#####                              |
| /##///## /##///## /##///## /###   /### /##///##                        /--- |
| /##  /## /##  /## /##  /## /#### /#### /##  ///          /------/     /     |
| /######/ /##  /## /##  /## /##/####/## //#####          /      /     /      |
| /##///## /##  /## /##  /## /##//##//##  /////##        /------/ __  /       |
| /##  /## /##  /## /##  /## /## /// /##      /##   __  /  __  / /  \/        |
| /##  /## /##  /## /##  /## /##     /## /##  /##  /  \/  /  \/  \__/         |
| /######/ //#####/ //#####/ /##     /## //#####/  \__/   \__/                |
| ///////   //////   //////  ///     ///  //////                              |
|                                                                             |
|                   Booms Object Oriented Music Structures.                   |
|Eli Barzilay:                                                   Maze is Life!|
+-----------------------------------------------------------------------------+

|#


;;;-------------------------------------------------------------------------;;;
;;;  N E T W O R K  file                                                    ;;;
;;;-------------------------------------------------------------------------;;;

(in-package :network)


;;;-------------------------------------------------------------------------;;;
;;;  U T I L I T I Y   definitions                                          ;;;
;;;-------------------------------------------------------------------------;;;

;;; Global constant.
(defconstant +NOT-CALCULATED+ :<not-calculated>)

;;; Default evaluating just returns a copy.
(defmethod EVALUATE ((x t))
  (make-copy x))

;;; Bind this variable to t for no error signals.
(defvar *NO-NETWORK-ERRORS* nil)


;;;-------------------------------------------------------------------------;;;
;;;  N O D E  class                                                         ;;;
;;;-------------------------------------------------------------------------;;;
;;; NODE is the network functional objects class.  It serves as a base class
;;; for all node classes in the system. The user never defines a node instance,
;;; he uses defnode to define a subclass for a node with all its extra
;;; information.
;;; Type-out is the output type value from the evaluate method of some node
;;; class.  Type-in is nil in case of no arguments (e.g. leaf), a list in the
;;; form of (T TYPE) in case of a list-arguments nodes where each argument has
;;; type TYPE and a list in the form of (n T1 T2 ... Tn) in case of a fixed
;;; n-arguments node where argument i has type Ti.

;;; The NODE class definition.
(defclass NODE ()
  ((type-in   :reader   type-in)
   (type-out  :reader   type-out)
   (arg-names :reader   arg-names)
   (uplinks   :accessor uplinks   :initform nil :initarg :uplinks)
   (arguments :reader   arguments :initform nil :initarg :arguments)
   (args-full-p-cache :accessor args-full-p-cache :initform +not-calculated+)))

(defmacro NODEP (obj)
  `(typep ,obj 'node))

;;; A standard printing routine for this class.
(defmethod PRINT-OBJECT ((n node) stream)
  (format stream "#<node:~a>" (type-of n)))

;;; Making a copy of a node - simply a new instance.
(defmethod MAKE-COPY ((n node))
  (make-instance (type-of n)))

;;; Get some information on the arguments: ARGS-NUM returns nil = no-args,
;;; n = n args, t = arg list.
(defmacro ARGS-NUM (node)
  `(and (type-in ,node) (first (type-in ,node))))

;;; Macros to determine the arguments number of a node.

(defmacro NO-ARGS-P (node)
  `(null (type-in ,node)))

(defmacro FIXED-ARGS-P (node)
  `(numberp (args-num ,node)))

(defmacro LIST-ARGS-P (node)
  `(eq (args-num ,node) t))

;;; Is the node ready for evaluation?
(defmethod ARGUMENTS-FULL ((n node))
  (when (eq (args-full-p-cache n) +not-calculated+)
    (setf (args-full-p-cache n)
          (and (not (member +no-value+ (arguments n)))
               (every #'(lambda (arg)
                          (or (not (nodep arg)) (arguments-full arg)))
                      (arguments n)))))
  (args-full-p-cache n))

;;; Making evaluate abort when there is some undefined arguments.
(defmethod EVALUATE :around ((n node))
  (cond ((arguments-full n) (call-next-method))
        (*no-network-errors* nil)
        (t (error "Undefined arguments, can't evaluate node ~a." n))))

;;; Is a given node evaluated?
(defmethod EVALUATED-P ((n node))
  (not (eql (evaluate-cache n) +not-calculated+)))

;;; Forgetting all values (caches), the real work will be done by methods that
;;; the defnode macro will create.
(defmethod FORGET-VALUE ((n node))
  (when (evaluated-p n)
    (mapc #'forget-value (uplinks n)))
  (setf (args-full-p-cache n) +not-calculated+))

;;; Dealing with the uplinks connecting.
(defmethod LINKUP ((down node) (up node))
  (setf (uplinks down) (adjoin up (uplinks down))))

;;; Forget values on a link.
(defmethod LINKUP :after ((down node) (up node))
  (forget-value up))

;;; Making sure no linkings of non-nodes, and no errors.
(defmethod LINKUP ((down t) (up t))
  t)

;;; Canceling an uplink - done after the argument has been removed from the
;;; parent node.
(defmethod UNLINKUP ((down node) (up node))
  (unless (member down (arguments up))
    (setf (uplinks down) (delete up (uplinks down)))))

;;; Making sure no unlinkings of non-nodes.
(defmethod UNLINKUP ((down t) (up t))
  t)

;;; This function is used to convert a node list to a list with corresponding
;;; node values - using the given function.

(defmethod GET-EVALS ((x t) (f function))
  (make-copy x))

(defmethod GET-EVALS ((s list) (f function))
  (mapcar #'(lambda (x) (get-evals x f)) s))

(defmethod GET-EVALS ((n node) (f function))
  (make-copy (funcall f n)))

;;; Check if a node is connected upwards to another node.
;;; The BFS iterator.
(defmethod CHECK-LINKED ((buffer list) (up node))
  (do ((head (pop buffer) (pop buffer)))
      ((or (null head) (eq up head)) (eq up head))
    (when (nodep head)
      (setf buffer (union buffer (uplinks head))))))

;;; The same iterator, looking for a list of targets.
(defmethod CHECK-LINKED ((buffer list) (ups list))
  (do ((head (pop buffer) (pop buffer)))
      ((or (null head) (member head ups)) (member head ups))
    (when (nodep head)
      (setf buffer (union buffer (uplinks head))))))

;;; Using a start node.
(defmethod CHECK-LINKED ((down node) (up node))
  (check-linked (list down) up))
(defmethod CHECK-LINKED ((down node) (up list))
  (check-linked (list down) up))

;;; Nothing is linked to non-nodes.
(defmethod CHECK-LINKED ((down t) (up t))
  nil)

;;; Type-out of non-nodes is their type.
(defmethod TYPE-OUT ((obj t))
  (type-of obj))

;;; Check a single object's type to be used as an argument.  The above type-out
;;; is actually not used for non-nodes - typep is better because it allows
;;; using Lisp compound types.
(defun CHECK-ARG-TYPE (what type)
  (or (eq what +no-value+)
      (and (nodep what) (subtypep (type-out what) type))
      (typep what type)))

;;; CHECK-ARGS checks if a node's arguments [or a given argument list] are
;;; valid for the node.
(defmethod CHECK-ARGS ((n node) &optional (args nil args-p))
  (let ((arguments (if args-p args (arguments n))))
    (and
     (not (check-linked n arguments))
     (cond
       ((fixed-args-p n)
        (and (= (length arguments) (length (rest (type-in n))))
             (every #'check-arg-type arguments (rest (type-in n)))))
       ((list-args-p n)
        (every #'(lambda (arg) (check-arg-type arg (second (type-in n))))
               arguments))
       (t t)))))

;;; CHECK-ARG checks if a specific argument can be set as some node's argument.
;;; The optional try-set argument will be true if the funcion should check that
;;; this value should have an available space now (meaningful only for fixed
;;; nodes).
(defmethod CHECK-ARG ((n node) (pos integer) (what t) &optional try-set)
  (cond
    ;; Prevent circles.
    ((check-linked n what) nil)
    ((fixed-args-p n) (and (< -1 pos (args-num n))
                           (check-arg-type what (nth pos (rest (type-in n))))
                           (or (not try-set)
                               ;; Position is free.
                               (eq (nth pos (arguments n)) +no-value+))))
    ((list-args-p n)  (and (< -1 pos (1+ (length (arguments n))))
                           (check-arg-type what (second (type-in n)))))
    ((no-args-p n) nil)))

;;; Set the argument list - first check, then disconnect/connect nodes and
;;; finally set the slot value.
(defmethod (setf ARGUMENTS) (newargs (n node))
  (cond
    ((check-args n newargs)
     (forget-value n)
     (let ((oldargs (slot-value n 'arguments)))
       (dolist (oldarg oldargs)
         (unless (member oldarg newargs)
           (unlinkup oldarg n)))
       (dolist (newarg newargs)
         (unless (member newarg oldargs)
           (linkup newarg n))))
     (setf (slot-value n 'arguments) newargs)
     t)
    (*no-network-errors* nil)
    (t (error "Wrong arguments type for node."))))

;;; Make sure that node instances have valid arguments, then link'em.
(defmethod INITIALIZE-INSTANCE ((n node) &rest initargs)
  (call-next-method)
  (when (and (arguments n) (type-in n) (not (check-args n)))
    (format t "~&Initial arguments type conflict - ignored!~%")
    (setf (slot-value n 'arguments)
          (if (list-args-p n)
            nil
            (make-list (args-num n) :initial-element +no-value+))))
  (if (listp (arguments n))
    (dolist (arg (arguments n)) (linkup arg n))
    (linkup (arguments n) n))
  n)


;;;-------------------------------------------------------------------------;;;
;;;  I N T E R F A C E  functions for node class                            ;;;
;;;-------------------------------------------------------------------------;;;

;;; NODE - Build a node using a given node class, possibly with arguments.
(defun NODE (nclass &rest initargs)
  (apply #'make-instance nclass initargs))

;;; Set a node using a given node class name, possibly with arguments.
(defmacro SETNODE (v nclass &rest initargs)
  `(setf ,v (make-instance ',nclass ,@initargs)))

;;; These three methods are used as an interface to the argument list of the
;;; node.  The "0" means these are program interface functions (start counting
;;; from 0)

;;; Get an argument.

(defmethod GETARG0 ((n node) (pos integer))
  (nth pos (arguments n)))

(defmethod GETARG0 :around ((n node) (pos integer))
  (cond ((and (not (no-args-p n)) (< -1 pos (length (arguments n))))
         (call-next-method))
        (*no-network-errors* nil)
        (t (error "GETARG0: No such location."))))

;;; Set an argument.

(defmethod SETARG0 ((n node) (pos integer) (what t))
  (cond
    ((fixed-args-p n) (setf (nth pos (arguments n)) what))
    ((list-args-p n)
     (if (zerop pos)
       (setf (slot-value n 'arguments) (cons what (arguments n)))
       (setf (rest (nthcdr (1- pos) (arguments n)))
             (cons what (nthcdr pos (arguments n)))))))
  (forget-value n)
  (linkup what n)
  :done)

(defmethod SETARG0 :around ((n node) (pos integer) (what t))
  (cond ((check-arg n pos what t) (call-next-method))
        (*no-network-errors* nil)
        (t (error "SETARG0: Error trying to set argument #~a of ~a to ~a"
                  pos n what))))

;;; Delete an argument.

(defmethod DELARG0 ((n node) (pos integer))
  (let ((temp (getarg0 n pos)))
    (cond
      ((fixed-args-p n)
       (setf (nth pos (arguments n)) +no-value+))
      ((list-args-p n)
       (if (zerop pos)
         (setf (slot-value n 'arguments) (rest (arguments n)))
         (setf (rest (nthcdr (1- pos) (arguments n)))
               (nthcdr (1+ pos) (arguments n))))))
    (unlinkup temp n))
  (forget-value n)
  :done)

(defmethod DELARG0 :around ((n node) (pos integer))
  (cond ((and (not (no-args-p n)) (< -1 pos (length (arguments n))))
         (call-next-method))
        (*no-network-errors* nil)
        (t (error "DELARG0: No such location."))))

;;; Human interface for the above - start counting from 1.

(defmacro GETARG (node pos)      `(getarg0 ,node (1- ,pos)))

(defmacro SETARG (node pos what) `(setarg0 ,node (1- ,pos) ,what))

(defmacro DELARG (node pos)      `(delarg0 ,node (1- ,pos)))

;;; This method will delete all connections of a given node, it is to be called
;;; just before a node is deleted.
(defmethod DELETE-LINKS ((n node))
  (labels ((positions (i x l)
             (cond ((null l) nil)
                   ((eql x (first l)) (cons i (positions (1+ i) x (rest l))))
                   (t (positions (1+ i) x (rest l))))))
    (dolist (p (reverse (uplinks n)))
      ;; Reverse will copy the uplinks since they'll be modified (by delarg0),
      ;; I don't remember why reverse and not copy-list...
      (dolist (pos (reverse (positions 0 n (arguments p))))
        ;; Here reverse is essential in case of deletions from an argument list
        (delarg0 p pos)))
    (let ((old-args (arguments n)))
      (setf (slot-value n 'arguments)
            (if (fixed-args-p n)
              (make-list (args-num n) :initial-element +no-value+)
              nil))
      (dolist (s old-args) (unlinkup s n))))
  :done)


;;;-------------------------------------------------------------------------;;;
;;;  D E F N O D E  macro                                                   ;;;
;;;-------------------------------------------------------------------------;;;
;;; This is the most important macro in the network system.  It is used to
;;; define a new node subclass to be used.  Its syntax is complex, because it
;;; defines all of the new class' details.  It builds the class, its input type
;;; list, its slots and its methods.  Syntax objects in double quotes are to be
;;; entered as is, and objects in '_'s are some simple argument (= symbol).
;;; Syntax: ( DEFNODE NAME ARG-LIST ANCESTORS SLOTS . METHODS )
;;;   NAME - the new class name
;;;   ARG-LIST - the input argument list can be none, fixed or list of args:
;;;       ::= "()" | "nil" | FIXED-ARGS | LIST-ARGS
;;;       FIXED-ARGS ::= ( { (_name_ _type_) }+ )
;;;       LIST-ARGS  ::= ( _name_ list _type_ )
;;;   ANCESTORS - the ancestors class list used for inheritance
;;;               if no node-subclass then node will be added as last one
;;;   SLOTS - slot declaration list for CLOS
;;;   METHODS - the list of methods, their parameters, and bodies:
;;;       ::= ( EVALUATE-M { METHOD }* )
;;;       METHOD ::= ( _name_ KEYWORDS := . BODY )
;;;       KEYWORDS ::= [[ ":cached" | ":fullargs" | ":lazy" |
;;;                    ":evaled" _eval-func_ ]]*
;;;       BODY - the method body, can use all of the node's formal arguments
;;;              and another argument 'self' which is the node itself.
;;;       EVALUATE-M - the first method which determine the node output type.
;;;           ::= ( "evaluate" _type_ := . BODY )
;;;   Method keyword explanations:
;;;     :cached   - cache method result
;;;     :fullargs - evaluate when all arguments are full (effective for fixed)
;;;     :evaled   - evaluate the arguments using their given method
;;;     :lazy     - use lazy evaluation - symbol-macrolet (effective for fixed)
;;;   The evaluation method type determines the node's type-out, it is always
;;;   cached and evaled using itself.  It is never lazy, but the defnode macro
;;;   can be easily modified to make it always lazy.

;;; Utility for displaying defnode syntax errors.
(defun DEFNODE-SYNTAX-ERROR (str &rest args)
  (error "Defnode syntax error: ~?" str args))

;;; The defnode macro.
(defmacro DEFNODE (name arguments ancestors slots eval-method &rest methods)
  ;; Check name.
  (unless (symbolp name)
    (defnode-syntax-error "Name=~a." name))
  ;; Check arguments.
  (unless (listp arguments)
    (defnode-syntax-error "Arguments=~a." arguments))
  (if (listp (first arguments))
    (dolist (argument arguments)
      (unless (and (listp argument)
                   (= 2 (length argument))
                   (symbolp (first argument))
                   (symbolp (second argument)))
        (defnode-syntax-error "Argument=~a." argument)))
    (unless (and (= (length arguments) 3)
                 (symbolp (first arguments))
                 (eql 'list (second arguments))
                 (symbolp (third arguments)))
      (defnode-syntax-error "Arguments=~a." arguments)))
  ;; Check ancestors.
  (unless (listp ancestors)
    (defnode-syntax-error "Ancestors=~a." ancestors))
  (dolist (ancestor ancestors)
    (unless (and (symbolp ancestor) (find-class ancestor nil))
      (defnode-syntax-error "~a is not a class." ancestor)))
  ;; Check slots.
  (unless (listp slots)
    (defnode-syntax-error "Slots=~a." slots))
  ;; Check evaluate method.
  (unless (and (listp eval-method)
               (> (length eval-method) 3)
               (eql (first eval-method) 'evaluate)
               (symbolp (second eval-method))
               (eql (third eval-method) :=))
    (defnode-syntax-error "Evaluate method = ~a." eval-method))
  ;; Check methods.
  (unless (listp methods)
    (defnode-syntax-error "Methods=~a." methods))
  (dolist (method methods)
    (unless (and (> (length method) 2)
                 (symbolp (first method))
                 (> (length (member := method)) 1))
      (defnode-syntax-error "Wrong method specification: ~a." method))
    (do* ((keyword (subseq method 1 (position := method)) (rest keyword)))
         ((null keyword))
      (cond
        ((eql (first keyword) :cached))
        ((eql (first keyword) :fullargs))
        ((eql (first keyword) :lazy))
        ((and (eql (first keyword) :evaled)
              (> (length keyword) 1))
         (pop keyword)
         (unless (symbolp (first keyword))
           (defnode-syntax-error "Evaled argument = ~a." (first keyword))))
        (t (defnode-syntax-error "~a is not a legal keyword."
                                 (first keyword))))))
  ;; Make sure node is a superclass.
  (setf ancestors (insure-superclass ancestors 'node))
  ;; Do the work.
  (let* ((o-type (second eval-method))
         (type-in (cond
                    ((null arguments) nil)
                    ((listp (first arguments))
                     (cons (length arguments) (mapcar #'second arguments)))
                    (t (list t (third arguments)))))
         (arg-names (cond
                      ((null arguments) nil)
                      ((listp (first arguments)) (mapcar #'first arguments))
                      (t (first arguments))))
         (m-list
          (cons (list* 'evaluate 'evaluate-cache 'evaluate nil nil
                       (nthcdr 3 eval-method))
                (make-method-descriptors methods))))
    `(progn
       (defclass ,name ,ancestors
         ((type-in   :allocation :class :initform ',type-in)
          (type-out  :allocation :class :initform ',o-type)
          (arg-names :allocation :class :initform ',arg-names)
          (arguments :initform
                     ,(if (or (null type-in) (eql (first type-in) t))
                        `nil
                        `(make-list ,(first type-in)
                                    :initial-element +no-value+)))
          ,@(make-cache-definitions m-list)
          ,@slots))
       ,@(make-node-method-definitions m-list name arguments type-in)
       (defmethod forget-value ((n ,name))
         (call-next-method)
         ,@(mapcar #'(lambda (m) `(setf (slot-value n ',m) +not-calculated+))
                   (remove-if #'null (mapcar #'second m-list))))
       (find-class ',name))))

;;; This function gets a list of method input as specified in the defnode macro
;;; syntax, and builds a list of method descriptors while checking the syntax.
;;; The generated list is:
;;;   (m-name m-cache-var m-eval-func m-full-args-p. m-body)
(defun MAKE-METHOD-DESCRIPTORS (methods)
  (mapcar
    #'(lambda (m)
        (let ((temp (list (first m) nil nil nil nil)))
          (do ((l (rest m)))
              ((equal (first l) :=)
               (nconc temp (rest l)))
            (case (pop l)
              (:cached
               (setf (second temp)
                     (symbol-append (first temp) '%%cache :network)))
              (:fullargs (setf (fourth temp) t))
              (:evaled   (setf (third  temp) (pop l)))
              (:lazy     (setf (fifth  temp) t))
              (t (defnode-syntax-error
                  "Unrecognized method keyword (~a)." (first m)))))))
    methods))

;;; This takes a method list as made above and produces the list of cache slots
;;; specification to be used in the class declaration.
(defun MAKE-CACHE-DEFINITIONS (m-list)
  (mapcar #'(lambda (s)
              `(,s :reader ,s :initform +not-calculated+))
          (remove-if #'null (mapcar #'second m-list))))

;;; This takes a method list as made above and produces the method definitions.
(defun MAKE-NODE-METHOD-DEFINITIONS (m-list name arguments type-in)
  (mapcar
    #'(lambda (m)
        (let* ((m-name      (first m))
               (cached      (second m))
               (evaled      (third m))
               (full-args-p (fourth m))
               (lazy-p      (fifth  m))
               (self-args (if evaled
                            `(get-evals (arguments self) #',evaled)
                            `(arguments self)))
               (m-body0 (nthcdr 5 m))
               (m-body1
                (cond
                  ((null type-in)
                   `(progn ,@m-body0))
                  ((eql (first type-in) t)
                   `(let ((,(first arguments) ,self-args)) ,@m-body0))
                  (lazy-p
                   `(symbol-macrolet
                        ,(let ((i -1))
                           (mapcar #'(lambda (arg)
                                       (list (first arg)
                                             `(nth ,(incf i) ,self-args)))
                                   arguments))
                      ,@m-body0))
                  (t
                   `(apply #'(lambda ,(mapcar #'first arguments) ,@m-body0)
                           ,self-args))))
               (m-body2
                (if cached
                  `(if (eq (,cached self) +not-calculated+)
                     (setf (slot-value self ',cached) ,m-body1)
                     (,cached self))
                  m-body1))
               (m-body3
                (if full-args-p
                  `(if (arguments-full self)
                     ,m-body2
                     +no-value+)
                  m-body2)))
          `(defmethod ,m-name ((self ,name)) ,m-body3)))
    m-list))


;;;-------------------------------------------------------------------------;;;
;;;  L E A F  class                                                         ;;;
;;;-------------------------------------------------------------------------;;;

;;; Leaf class declaration - a constant node.
(defclass LEAF (node)
  ((type-in   :allocation :class    :initform nil)
   (type-out  :allocation :instance :initarg :type-out)
   (arguments :reader evaluate)))

(defmacro LEAFP (obj)
  `(typep ,obj 'leaf))

;;; A standard printing routine for this class.
(defmethod PRINT-OBJECT ((l leaf) stream)
  (format stream "#<leaf:~a=~a>" (type-out l) (arguments l)))

;;; Making a copy of a node - a new instance with a copy of the value.
(defmethod MAKE-COPY ((l leaf))
  (make-instance (type-of l) :type-out  (type-out l)
                             :arguments (make-copy (arguments l))))

;;; Redefined - this node is ready when its argument is not a no-value.
(defmethod ARGUMENTS-FULL ((l leaf))
  (not (eql (arguments l) +no-value+)))

;;; This is needed to always return t for the forget-value around :method to
;;; allow the operation.
(defmethod EVALUATED-P ((n leaf))
  t)

;;; Check possible argument.  Not supplying args is actually meaningles for a
;;; leaf.
(defmethod CHECK-ARGS ((l leaf) &optional (args nil args-p))
  (let ((arguments (if args-p args (arguments l))))
    (check-arg-type arguments (type-out l))))

;;; The check for setting a new argument.
(defmethod (setf ARGUMENTS) (newval (l leaf))
  (cond ((check-args l newval)
         (forget-value l)
         (setf (slot-value l 'arguments) newval)
         t)
        (*no-network-errors* nil)
        (t (error "Can't change the leaf type."))))

;;; These methods are not for leaves.

(defmethod GETARG0 :around ((l leaf) (pos integer))
  (error "GETARG0: Cannot use with a leaf"))

(defmethod SETARG0 :around ((l leaf) (pos integer) (what t))
  (error "SETARG0: Cannot use with a leaf"))

(defmethod DELARG0 :around ((l leaf) (pos integer))
  (error "DELARG0: Cannot use with a leaf"))

;;; This is copied here because we don't want to reset the arguments list.
(defmethod DELETE-LINKS ((l leaf))
  (labels ((positions (i x l)
             (cond ((null l) nil)
                   ((eql x (first l)) (cons i (positions (1+ i) x (rest l))))
                   (t (positions (1+ i) x (rest l))))))
    (dolist (p (reverse (uplinks l)))
      (dolist (pos (reverse (positions 0 l (arguments p))))
        (delarg0 p pos)))
    (let ((old-arg (arguments l)))
      (setf (slot-value l 'arguments) +no-value+)
      (unlinkup old-arg l)))
  :done)

;;; Build a leaf, given a value.
(defun LEAF (val &rest keyword-args)
  (let ((l (apply #'make-instance 'leaf :type-out (type-of val) keyword-args)))
    (setf (slot-value l 'arguments) val)
    l))

;;; Set a leaf, given a value.
(defmacro SETLEAF (v val &rest keyword-args)
  `(setf ,v (leaf ,val ,@keyword-args)))


;;;-------------------------------------------------------------------------;;;
;;;  C O P Y - N O D E S  definition                                        ;;;
;;;-------------------------------------------------------------------------;;;
;;; This will copy a list of nodes - the whole list is copied with non-node
;;; connections, node connections are copied along the node copies if inside
;;; the list, otherwise deleted.  If a node/argss list is supplied, it is
;;; destructively modified to contain the node copies (this list is a result of
;;; abstraction-args below).

(defun COPY-NODES (old-nodes &optional node/argss)
  (let ((new-nodes (make-copy old-nodes))
        (old--new  (make-hash-table)))
    ;; Copy all argument lists and fill the hash table.
    (do ((old old-nodes (rest old))
         (new new-nodes (rest new)))
        ((null old))
      (setf (gethash (first old) old--new) (first new)))
    ;; Replace every node argument with its new copy.
    (do ((old old-nodes (rest old))
         (new new-nodes (rest new)))
        ((null old))
      (when (and (nodep (first new)) (not (leafp (first new))))
        (let ((new-args nil))
          (dolist (arg (arguments (first old)))
            (cond
              ((gethash arg old--new) (push (gethash arg old--new) new-args))
              ((not (nodep arg))      (push (make-copy arg)        new-args))
              ((or (fixed-args-p (first new)) node/argss)
                                      (push +no-value+             new-args))))
          (setf (arguments (first new)) (nreverse new-args)))))
    (when node/argss
      (dolist (node/args node/argss)
        (dolist (node/arg node/args)
          (setf (first node/arg) (gethash (first node/arg) old--new)))))
    new-nodes))


;;;-------------------------------------------------------------------------;;;
;;;  A B S T R A C T I O N  class                                           ;;;
;;;-------------------------------------------------------------------------;;;

;;; The abstraction class holds an internal node list, and its root, and a
;;; node/argss list for setting arguments.  The node/argss is a list with an
;;; element for each argument - each element is a list of pairs for the same
;;; argument - the inner node and the position the argument.
(defclass ABSTRACTION (node)
  ((root            :initarg :root           :accessor root-node)
   (internal-nodes  :initarg :internal-nodes :accessor internal-nodes)
   (node/argss      :initarg :node/argss     :accessor node/argss)
   (type-in         :initarg :type-in        :accessor type-in :initform nil)
   (arg-names       :initarg :arg-names      :accessor arg-names)))

(defmacro ABSTRACTIONP (obj)
  `(typep ,obj 'abstraction))

;;; These methods make abstractions behave like their root node.

(defmethod TYPE-OUT    ((a abstraction)) (type-out    (root-node a)))

(defmethod EVALUATE    ((a abstraction)) (evaluate    (root-node a)))

(defmethod EVALUATED-P ((a abstraction)) (evaluated-p (root-node a)))

;;; Forget-value should be done although it is not evaluated - the inner nodes
;;; gets a forget value, and they don't pump it outside the abstraction - it
;;; does not have a value but should pump the message up.
(defmethod FORGET-VALUE ((a abstraction))
  (mapc #'forget-value (uplinks a))
  (setf (args-full-p-cache a) +not-calculated+))

;;; If the arguments are setf'ed - set the internal node arguments.
(defmethod (setf ARGUMENTS) (newargs (a abstraction))
  (if (call-next-method)
    (do ((node/argss (node/argss a) (rest node/argss))
         (args       (arguments  a) (rest args)))
        ((null args) t)
      (dolist (node/arg (first node/argss))
        (unless (eq (getarg0 (first node/arg) (second node/arg)) (first args))
          (delarg0 (first node/arg) (second node/arg))
          (setarg0 (first node/arg) (second node/arg) (first args)))))
    nil))

;;; Setting an argument sets the inner node arguments also.
(defmethod SETARG0 ((a abstraction) (pos integer) (what t))
  (if (call-next-method)
    (let ((node/args (nth pos (node/argss a))))
      (dolist (node/arg node/args)
        (delarg0 (first node/arg) (second node/arg))
        (setarg0 (first node/arg) (second node/arg) what))
      :done)
    nil))

;;; Deleting an argument stores a +no-value+ instead.
(defmethod DELARG0 ((a abstraction) (pos integer))
  (if (call-next-method)
    (let ((node/args (nth pos (node/argss a))))
      (dolist (node/arg node/args)
        (delarg0 (first node/arg) (second node/arg))
        (setarg0 (first node/arg) (second node/arg) +no-value+))
      :done)
    nil))

;;; Dlete-links should delete the links to the internal argument nodes.
;;; The variable is for preventing this deletion in case of an abstraction
;;; explosion.

(defvar *ABSTRACTION-EXPLODING* nil)

(defmethod DELETE-LINKS ((a abstraction))
  (unless *abstraction-exploding*
    (dolist (node/args (reverse (node/argss a)))
      (dolist (node/arg (reverse node/args))
        (delarg0 (first node/arg) (second node/arg)))))
  (call-next-method)
  :done)

;;; Making a copy of an abstraction.
(defmethod MAKE-COPY ((a abstraction))
  (let* ((result (call-next-method))
         (node/argss+root
          (cons (list (list (root-node a) nil))
                (mapcar #'(lambda (node/args) (mapcar #'copy-list node/args))
                        (node/argss a))))
         (new-nodes (copy-nodes (internal-nodes a) node/argss+root)))
    (setf (root-node      result) (caaar node/argss+root)
          (internal-nodes result) new-nodes
          (node/argss     result) (rest node/argss+root)
          (type-in        result) (type-in a)
          (arg-names      result) (arg-names a)
          (arguments      result) (make-list (args-num result)
                                             :initial-element +no-value+))
    result))


;;;-------------------------------------------------------------------------;;;
;;;  A B S T R A C T I O N  creation                                        ;;;
;;;-------------------------------------------------------------------------;;;

;;; Create an abstraction - if possible.  An abstraction is actually a group
;;; of nodes, minimized to a single abstraction node.  This node holds all
;;; other nodes within itself, and it has an arugment for each node that was
;;; outside the original node group.  This abstraction node is already
;;; connected like the original node group.  The original group should have a
;;; single root node, and it should not have a node with output connecting to a
;;; node outside the group (except for he root).
(defun MAKE-ABSTRACTION (node-list same-node-same-arg &optional abs-node)
  (let ((root (find-root-node node-list)))
    (when root
      (let* ((node/argss (abstraction-args root node-list same-node-same-arg))
             (type-in    (cons (length node/argss)
                               (mapcar
                                #'(lambda (nas)
                                    (let ((na (first nas)))
                                      (if (list-args-p (first na))
                                        (second (type-in (first na)))
                                        (nth (second na)
                                             (rest (type-in (first na)))))))
                                node/argss)))
             (arg-names  (let ((l nil) (n (length node/argss)))
                           (dotimes (i n l)
                             (push (intern (format nil "~:@(arg-~R~)" (- n i))
                                           :keyword)
                                   l))))
             (arguments  (mapcar #'(lambda (nas)
                                     (getarg0 (first (first nas))
                                              (second (first nas))))
                                 node/argss))
             (result     (or abs-node (make-instance 'abstraction))))
        (setf (root-node      result) root
              (internal-nodes result) node-list
              (node/argss     result) node/argss
              (type-in        result) type-in
              (arg-names      result) arg-names
              (arguments      result) arguments)
        (dolist (up (uplinks root))
          (replace-args up root result))
        result))))

;;; Explode an abstraction - disconnect the abstraction and reconnect the root
;;; node (other nodes are already connected).  Return the list of exploded
;;; internal nodes.
(defun EXPLODE-ABSTRACTION (node)
  (dolist (up (uplinks node))
    (replace-args up node (root-node node)))
  (dolist (node/args (reverse (node/argss node)))
    (dolist (node/arg (reverse node/args))
      (when (eq (getarg0 (first node/arg) (second node/arg)) +no-value+)
        (delarg0 (first node/arg) (second node/arg)))))
  (let ((*abstraction-exploding* t))
    (delete-links node))
  (internal-nodes node))

;;; Return the root node of a given node list or nil if none.  There shouldn't
;;; be a node with no uplinks except the root node.  If a non-root node has an
;;; uplink outside the list - then nil is returned.
(defun FIND-ROOT-NODE (nodes)
  (let ((root      nil)
        (all-nodes nil))
    ;; Calculate the closure of all nodes in nodes (if some are abstractions).
    (do* ((nodes nodes)
          (node  (pop nodes) (pop nodes)))
        ((null node))
      (when (abstractionp node)
        (setf nodes (append (internal-nodes node) nodes)))
      (push node all-nodes))
    ;; Do the check.
    (dolist (node nodes root)
      (cond
        ((not (nodep node))
         (return-from find-root-node nil))
        (root
         (when (or (null (uplinks node))
                   (notevery #'(lambda (n) (member n all-nodes))
                             (uplinks node)))
           (return-from find-root-node nil)))
        ((or (null (uplinks node))
             (notany #'(lambda (n) (member n all-nodes)) (uplinks node)))
         (setf root node))
        ((notevery #'(lambda (n) (member n all-nodes)) (uplinks node))
         (return-from find-root-node nil))))))

;;; Will this node-list contain arguments when converted to an abstraction?
;;; Used for a decision weather a leaf abstraction should be built.
(defun HAVE-ABSTRACT-ARGS-P (node-list)
  (dolist (node node-list)
    (unless (nodep node)
      (return-from have-abstract-args-p nil))
    (unless (or (leafp node)
                (every #'(lambda (arg)
                           (if (nodep arg)
                             (member arg node-list)
                             (not (eq arg +no-value+))))
                       (arguments node)))
      (return-from have-abstract-args-p t)))
  nil)

;;; Return a list of an abstraction arguments - node objects not in the node
;;; list, and +no-values+.  Each element is a two valued list - the node and
;;; the number of argumet.  The first element will be of the root node and nil.
(defun ABSTRACTION-ARGS (root node-list same-node-same-arg)
  (let ((result  nil)
        (scanned nil))
    (labels ((scan-tree (node)
               (unless (or (leafp node) (member node scanned))
                 (push node scanned)
                 (let ((i 0))
                   (dolist (arg (arguments node))
                     (cond
                       ((eq arg +no-value+)
                        (push (list (list node i)) result))
                       ((nodep arg)
                        (if (member arg node-list)
                          (scan-tree arg)
                          (if same-node-same-arg
                            (let ((found nil))
                              (dolist (nas result)
                                (when (eq arg
                                          (getarg0 (first (first nas))
                                                   (second (first nas))))
                                  (setf found nas)))
                              (if found
                                (nconc found (list (list node i)))
                                (push (list (list node i)) result)))
                            (push (list (list node i)) result)))))
                     (incf i))))))
      (scan-tree root)
      (nreverse result))))

;;; This will replace all arguments of node that are from-node to to-node.
(defun REPLACE-ARGS (node from-node to-node)
  (dotimes (i (length (arguments node)))
    (when (eql (getarg0 node i) from-node)
      (delarg0 node i)
      (setarg0 node i to-node))))


;;;-------------------------------------------------------------------------;;;
;;;  I D E N T I T Y  class                                                 ;;;
;;;-------------------------------------------------------------------------;;;

;;; This class is used as a single-argument fixed node that will output
;;; anything that is input to it.  This can be used for having a single
;;; abstraction argument connected to several internal nodes.
(defclass IDENTITY (node)
  ((arguments  :initform (list +no-value+))
   (type-in    :allocation :class :initform nil)
   (type-out   :allocation :class :initform nil)
   (arg-names  :allocation :class :initform '(input))))

;;; This will find the minimum type of a given node's uplinks.
;;; WARNING: This assumes that the node types include a minimal type.
(defmethod MIN-UPLINKS-TYPE ((n node))
  (let ((result t))
    ;; Find the minimum type.
    (dolist (up (uplinks n))
      (cond
        ((typep up 'identity)
         (let ((up-min-type (min-uplinks-type up)))
           (when (subtypep up-min-type result)
             (setf result up-min-type))))
        ((list-args-p up)
         (when (subtypep (second (type-in up)) result)
           (setf result (second (type-in up)))))
        (t
         (do ((args  (arguments up)      (rest args))
              (types (rest (type-in up)) (rest types)))
             ((null args))
           (when (and (eq (first args) n)
                      (subtypep (first types) result))
             (setf result (first types)))))))
    result))

;;; The input type is determined by an existing argument or uplink.  Do not use
;;; getarg0 because it uses type-in.  This will be called many times, so keep
;;; efficiency - have a single return cons cell.
(defmethod TYPE-IN ((i identity))
  (let ((result '(1 t)))
    (setf (second result)
          (if (arguments-full i)
            (type-out (first (arguments i)))
            (min-uplinks-type i)))
    result))

;;; The output type is exactly the same as the first argument input type.
(defmethod TYPE-OUT ((i identity))
  (let ((result (second (type-in i))))
    (and (not (eq result t)) result)))

;;; Make identity evaluate as its argument.
(defmethod EVALUATE ((i identity))
  (evaluate (first (arguments i))))

;;; Is the identity evaluated? - Answer this question using the input argument,
;;; this requiers a check that the arguments are bound and full (when loading,
;;; forget-value will call this before the arguments slot is bound).
(defmethod EVALUATED-P ((i identity))
  (and (slot-boundp i 'arguments)
       (arguments-full i)
       (or (not (nodep (first (arguments i))))
           (evaluated-p (first (arguments i))))))

;;; Making an identity copy.
(defmethod MAKE-COPY ((i identity))
  (make-instance (type-of i)))


;;;-------------------------------------------------------------------------;;;
;;;  L I S P - V A L U E  class                                             ;;;
;;;-------------------------------------------------------------------------;;;

;;; This is a leaf object with value that is the minimum of all its parents.
;;; It is usefull for having a general value that can be changed according to
;;; what it is connected to.
(defclass LEAF-VALUE (leaf) ())

;;; The output type is the type of the stored value.  This is not used for
;;; argument checkings - see below.
(defmethod TYPE-OUT ((lv leaf-value))
  (type-of (arguments lv)))

;;; Checking arguments is done for the minimum parent type.  In case there are
;;; no uplinks, this will be t which allows any value.
(defmethod CHECK-ARGS ((lv leaf-value) &optional (args nil args-p))
  (let ((arguments (if args-p args (arguments lv))))
    (check-arg-type arguments (min-uplinks-type lv))))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

;;; Initialize network variables.
(defun INITIALIZE-NETWORK ()
  (setf *no-network-errors* nil))


;;;-------------------------------------------------------------------------;;;
;;;  D O N E !                                                              ;;;
;;;-------------------------------------------------------------------------;;;

(provide :NETWORK)

